# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNUGeneral Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

from __future__ import division

import math

import numpy as np

###

def radius_and_pressure(N, R, Es):
    """
    Given normal load, sphere radius and contact modulus compute contact radius
    and peak pressure.

    Parameters
    ----------
    N : float
        Normal force.
    R : float
        Sphere radius.
    Es : float
        Contact modulus: Es = E/(1-nu**2) with Young's modulus E and Poisson
        number nu.

    Returns
    -------
    a : float
        Contact radius.
    p0 : float
        Maximum pressure inside the contacting area (right under the apex).
    """

    a = R*(3./4*( N/(Es*R**2) ))**(1./3)
    p0 = 3*N/(2*math.pi*a*a)
    
    return a, p0


def surface_stress(r, poisson=0.5):
    """
    Given distance from the center of the sphere, contact radius and Poisson
    number contact, compute the stress at the surface.

    Parameters
    ----------
    r : array_like
        Array of distance (from the center of the sphere in units of contact
        radius a).
    poisson : float
        Poisson number.

    Returns
    -------
    pz : array
        Contact pressure (in units of maximum pressure p0).
    sr : array
        Radial stress (in units of maximum pressure p0).
    stheta : array
        Azimuthal stress (in units of maximum pressure p0).
    """

    mask0 = np.abs(r) < 1e-6
    maski = np.logical_and(r < 1.0, np.logical_not(mask0))
    masko = np.logical_and(np.logical_not(maski), np.logical_not(mask0))
    r_0 = r[mask0]
    r_i = r[maski]
    r_o = r[masko]
    
    # Initialize
    pz = np.zeros_like(r)
    sr = np.zeros_like(r)
    stheta = np.zeros_like(r)
    
    # Solution at r=0
    if mask0.sum() > 0:
        pz[mask0] = np.ones_like(r_0)
        sr[mask0] = -(1.+2*poisson)/2.*np.ones_like(r_0)
        stheta[mask0] = -(1.+2*poisson)/2.*np.ones_like(r_0)

    # Solution inside the contact radius
    if maski.sum() > 0:
        r_a_sq = r_i**2
        pz[maski] = np.sqrt(1-r_a_sq)
        sr[maski] = (1.-2.*poisson)/(3.*r_a_sq)*(1.-(1.-r_a_sq)**(3./2))- \
            np.sqrt(1.-r_a_sq)
        stheta[maski] = -(1.-2.*poisson)/(3.*r_a_sq)*(1.-(1.-r_a_sq)**(3./2))- \
            2*poisson*np.sqrt(1.-r_a_sq)
    
    # Solution outside of the contact radius
    if masko.sum() > 0:
        r_a_sq = r_o**2
        po = (1.-2.*poisson)/(3.*r_a_sq)
        sr[masko] = po
        stheta[masko] = -po
    
    return pz, sr, stheta


def centerline_stress(z, poisson=0.5):
    """
    Given distance from the center of the sphere, contact radius and Poisson
    number contact, compute the stress at the surface.

    Parameters
    ----------
    z : array_like
        Array of depths (from the center of the sphere in units of contact
        radius a).
    poisson : float
        Poisson number.

    Returns
    -------
    srr : array
        Radial stress (in units of maximum pressure p0).
    szz : array
        Contact pressure (in units of maximum pressure p0).
    """

    srr = -(1.+poisson)*(1.-z*np.arctan(1./z)) + 1./(2.*(1.+z**2))
    szz = -1./(1.+z**2)

    return srr, szz

    
def surface_displacements(r):
    """
    Return the displacements at the surface due to an indenting sphere.
    See: K.L. Johnson, Contact Mechanics, p. 61

    Parameters
    ----------
    r : array_like
        Radial position normalized by contact radius a.

    Returns
    -------
    uz : array
        Normal displacements at the surface of the contact (in units of
        p0/Es * a where p0 is maximum pressure, Es contact modulus and a 
        contact radius).
    """

    maski = r < 1.0
    masko = np.logical_not(maski)
    r_i = r[maski]
    r_o = r[masko]
    
    # Initialize
    uz = np.zeros_like(r)
    
    # Solution inside the contact circle
    if maski.sum() > 0:
        uz[maski] = -math.pi*(2.-r_i**2)/4.

    # Solution outside the contact circle
    if masko.sum() > 0:
        uz[masko] = (-(2.-r_o**2)*np.arcsin(1./r_o) - 
            r_o*np.sqrt(1.-(1./r_o)**2))/2.

    return uz


def stress(r, z, poisson=0.5):
    """
    Return components of the stress tensor in the interior of the Hertz solid.
    This is the solution given by: M.T. Huber, Ann. Phys. 319, 153 (1904)

    Note that the stress tensor at any point in the solid has the form below in
    a cylindrical coordinate system centered at the tip apex. Some off-diagonal
    components are zero by rotational symmetry. stt is the circumferential
    component, srr the radial component and szz the normal component of the
    stress tensor.

            / stt  0   0  \ 
        s = |  0  srr srz |
            \  0  srz szz /

    Parameters
    ----------
    r : array_like
        Radial position (in units of the contact radius a).
    z : array_like
        Depth (in units of the contact radius a).
    poisson: float
        Poisson number.

    Returns
    -------
    stt : array
        Circumferential component of the stress tensor (in units of maximum
        pressure p0).
    srr : array
        Radial component of the stress tensor (in units of maximum pressure p0).
    szz : array
        Normal component of the stress tensor (in units of maximum pressure p0).
    srz : array
        Shear component of the stress tensor (in units of maximum pressure p0).
    """

    p = r**2+z**2-1
    u = p/2 + np.sqrt(p**2/4+z**2)
    sqrtu = np.sqrt(u)

    # Note: The equation for u is
    #     r**2/(1+u) + z**2/u = 1
    # With r' = r/sqrt(1+u) and z' = z/sqrt(u) this expression becomes
    #     r'**2 + z'**2 = 1, hence z' = sqrt(1-r'**2)
    # Variables r and z below are substituted by r->r' and z->z'.
    r = r/np.sqrt(1+u)
    z = np.sqrt(1-r**2) # equiv. to z=u/sqrt(u), but defined for z=0

    # Precompute arctan
    sqrtu_arctan_inv_sqrtu = sqrtu*np.arctan(1./sqrtu)

    # The next two expressions give numerical problems at the tip center and the
    # contact edge, regularize with the asymptotic value.

    # Note: (1-z**3)/r**2->3/2 for r->0
    one_minus_z3_div_r2 = 3./2.*np.ones_like(r)
    mask = r > 0
    one_minus_z3_div_r2[mask] = (1.-z[mask]**3)/r[mask]**2

    # Note: z**2/(u+z**2)->1 for u+z**2->0
    u_plus_z2 = u+z**2
    z2_div_u_plus_z2 = np.where(u_plus_z2 > 0., z**2/u_plus_z2, np.ones_like(z))

    # Compute stresses
    # Note: the factor 1/(1+u) stems from the substitution r->r' and z->z' above
    stt = (1.-2.*poisson)/3. * 1./(1.+u) * one_minus_z3_div_r2 + \
        z*(2.*poisson+ (1.-poisson)*u/(1.+u) - (1.+poisson)*sqrtu_arctan_inv_sqrtu)
    szz = z*z2_div_u_plus_z2
    srr = -( (1.-2.*poisson)/3. * 1./(1+u) * one_minus_z3_div_r2 + \
         z*z2_div_u_plus_z2 + \
         z*((1.-poisson)*u/(1.+u) + (1.+poisson)*sqrtu_arctan_inv_sqrtu - 2.) )
    srz = r*z2_div_u_plus_z2 * sqrtu/np.sqrt(1.+u)

    return -stt, -srr, -szz, -srz


def stress_Cartesian(x, y, z, poisson=0.5):
    """
    Return components of the stress tensor in the interior of solid due to
    normal Hertz loading.
    This is the solution given by:
    G.M. Hamilton, Proc. Instn. Mech. Engrs. 197C, 53-59 (1983)

    Parameters
    ----------
    x, y : array_like
        In-plane positions (in units of the contact radius a).
    z : array_like
        Depth (in units of the contact radius a).
    poisson : float
        Poisson number.

    Returns
    -------
    sxx, syy, szz, syz, sxz, sxy : array
        Individual components of the Cartesian stress tensor.
    """

    def stress_offcenter(x, y, z, r_sq, poisson=0.5):
        A = r_sq + z**2 - 1
        S = np.sqrt(A**2 + 4*z**2)

        M = np.sqrt((S+A)/2)
        N = np.sqrt((S-A)/2)
        phi = np.arctan2(1, M)

        G = M**2 - N**2 + z*M - N
        H = 2*M*N + M + z*N

        sxx = (1+poisson)*z*phi+1/r_sq*((y**2-x**2)/r_sq*((1-poisson)*N*z**2-(1-2*poisson)/3*(N*S+2*A*N+1)-poisson*M*z)-N*(x**2+2*poisson*y**2)-M*x**2*z/S)
        syy = (1+poisson)*z*phi+1/r_sq*((x**2-y**2)/r_sq*((1-poisson)*N*z**2-(1-2*poisson)/3*(N*S+2*A*N+1)-poisson*M*z)-N*(y**2+2*poisson*x**2)-M*y**2*z/S)
        szz = -N+z*M/S
        sxy = x*y*(1-2*poisson)/r_sq**2*(-N*r_sq+2/3*N*(S+2*A)-z*(z*N+M)+2/3)+x*y*z/r_sq**2*(-M*r_sq/S-z*N+M)
        syz = -z*(y*N/S-y*z*H/(G**2+H**2))
        sxz = -z*(x*N/S-x*z*H/(G**2+H**2))

        return sxx, syy, szz, syz, sxz, sxy

    x = np.asarray(x)
    y = np.asarray(y)
    z = np.asarray(z)

    sxx = np.zeros_like(x)
    syy = np.zeros_like(x)
    szz = np.zeros_like(x)
    syz = np.zeros_like(x)
    sxz = np.zeros_like(x)
    sxy = np.zeros_like(x)

    r_sq = x**2 + y**2

    mask = r_sq > 0
    xx, yy, zz, yz, xz, xy = stress_offcenter(x[mask], y[mask], z[mask],
                                              r_sq[mask], poisson=poisson)
    sxx[mask] = xx
    syy[mask] = yy
    szz[mask] = zz
    syz[mask] = yz
    sxz[mask] = xz
    sxy[mask] = xy

    mask = np.logical_not(mask)
    if mask.sum() > 0:
        z = z[mask]
        sxx[mask] = ((1+poisson)*(z*np.arctan2(1, z)-1)+1/(2*(1+z**2)))
        syy[mask] = sxx[mask]
        szz[mask] = -1/(1+z**2)

    return sxx, syy, szz, syz, sxz, sxy


def stress_for_tangential_loading(x, y, z, poisson=0.5):
    """
    Return components of the stress tensor in the interior of solid due to
    tangential (Hertz) loading.
    This is the solution given by:
    G.M. Hamilton, Proc. Instn. Mech. Engrs. 197C, 53-59 (1983)

    Parameters
    ----------
    x, y : array_like
        In-plane positions (in units of the contact radius a).
    z : array_like
        Depth (in units of the contact radius a).
    poisson : float
        Poisson number.

    Returns
    -------
    sxx, syy, szz, syz, sxz, sxy : array
        Individual components of the Cartesian stress tensor.
    """

    def stress_offcenter(x, y, z, r_sq, poisson=0.5):
        A = r_sq + z**2 - 1
        S = np.sqrt(A**2 + 4*z**2)

        M = np.sqrt((S+A)/2)
        N = np.sqrt((S-A)/2)
        phi = np.arctan2(1, M)

        G = M**2 - N**2 + z*M - N
        H = 2*M*N + M + z*N

        sxx = -x*(poisson/4+1)*phi+x*M/r_sq**2*((3/2-2*x**2/r_sq)*(S*poisson-2*A*poisson+z**2)+x**2*z**2/S+7*poisson*r_sq/4-2*poisson*x**2+r_sq)+x*z*N/r_sq**2*((3/2-2*x**2/r_sq)*(-S/6*(1-2*poisson)-A/3*(1-2*poisson)-1/2*(z**2+3))+x**2/S-poisson*r_sq/4-7*r_sq/4)+4*x*z/(3*r_sq**2)*(3/2-2*x**2/r_sq)*(1-2*poisson)
        syy = -3*poisson*x*phi/4+x*M/r_sq**2*((1/2-2*y**2/r_sq)*(poisson*(S-2*A+r_sq)+z**2)+y**2*z**2/S+3/4*poisson*r_sq)+z*x*N/r_sq**2*((1/2-2*y**2/r_sq)*(-S/6*(1-2*poisson)-A/3*(1-2*poisson)-z**2/2-3/2)+y**2/S-3/4*poisson*r_sq-r_sq/4)+4/3*z*x/r_sq**2*(1/2-2*y**2/r_sq)*(1-2*poisson)
        szz = z*x*N/(2*r_sq)*(1-(r_sq+z**2+1)/S)
        sxy = y/2*(poisson/2-1)*phi+y*M/r_sq**2*(x**2*z**2/S+poisson*((S-2*A)*(1/2-2*x**2/r_sq)-2*x**2+r_sq/4)+r_sq/2+z**2*(1/2-2*x**2/r_sq))+y*z*N/r_sq**2*((1/2-2*x**2/r_sq)*((2*poisson-1)*(S/6+A/3)-z**2/2-3/2-r_sq/2)+r_sq*poisson/4+x**2/S-y**2/2-3*x**2/2)+4*y*z/(3*r_sq**2)*(1/2-2*x**2/r_sq)*(1-2*poisson)
        syz = x*y*z/(2*r_sq**2)*(M*(1/2+1/S*(z**2/2-3/2-r_sq/2))+z*N/2*(-3+1/S*(5+z**2+r_sq)))
        sxz = 3*z*phi/2+z*M/r_sq*(1+x**2/r_sq-x**2/S)+N/r_sq*(-3/4*(S+2*A)+z**2-3/4-1/4*r_sq+z**2/2*(1/2-2*x**2/r_sq))

        return sxx, syy, szz, syz, sxz, sxy

    x = np.asarray(x)
    y = np.asarray(y)
    z = np.asarray(z)

    sxx = np.zeros_like(x)
    syy = np.zeros_like(x)
    szz = np.zeros_like(x)
    syz = np.zeros_like(x)
    sxz = np.zeros_like(x)
    sxy = np.zeros_like(x)

    r_sq = x**2 + y**2

    mask = r_sq > 0
    xx, yy, zz, yz, xz, xy = stress_offcenter(x[mask], y[mask], z[mask],
                                              r_sq[mask], poisson=poisson)
    sxx[mask] = xx
    syy[mask] = yy
    szz[mask] = zz
    syz[mask] = yz
    sxz[mask] = xz
    sxy[mask] = xy

    mask = np.logical_not(mask)
    if mask.sum() > 0:
        z = z[mask]
        sxz[mask] += -1+3/2*z*np.arctan2(1, z)-z**2/(2*(1+z**2))

    return sxx, syy, szz, syz, sxz, sxy




# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#                  Adrien Gola, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# ======================================================================


"""
 Usage : python plot.py [filename]
 
 This tool script allow to quicly plot data from text files arranged in column with space as separator. It uses a GUI through Tkinter
 "log.lammp" can be read and splited according to different "run" present in the log file
                                                                                            
 Adrien Gola // 4.11.2015                                                                        
"""

from __future__ import print_function

import numpy as np
import matplotlib
matplotlib.rcParams['backend'] = "Qt4Agg"
import matplotlib.pyplot as p 

import os,sys
import os.path

import Tkinter as tk
import ttk

# ------------------------------------
# ------- Function definition --------
# ------------------------------------
def close_windows():
    root.destroy()

def plot_button():
    global xlabel,ylabel
    x = d[:,Xvar.get()]
    y = d[:,Yvar.get()]
  
    if xlabel.get() != "X-axis label":
        hx = xlabel.get()
    else:
        hx = h[Xvar.get()]
    if ylabel.get() != "Y-axis label":
        hy = ylabel.get()
    else:
        hy = h[Yvar.get()]
  
    p.plot(x, y, 'o')
    p.xlabel(hx)
    p.ylabel(hy)
    p.show() 
  
def process_file():
    global variables_frame,h,d,Xvar,Yvar,skipline,flag_loglammps
    selected_file = sorted(outlogfiles)[Fvar.get()]
    if flag_loglammps:
        h = open(selected_file,'r').readlines()[0].strip('#').strip().split(" ") # headers list 
        d = np.loadtxt(selected_file, skiprows=1) # data  
    else:
        h = filter(None,open(selected_file,'r').readlines()[0+int(skipline.get())].strip('#').strip().split(" ")) # headers list 
        d = np.loadtxt(selected_file, skiprows=1+int(skipline.get())) # data   
    try:
        for child in variables_frame.winfo_children():
            child.destroy()
    except:
        pass
    #Xvar
    tk.Label(variables_frame,text="Select X-axis variable to plot",anchor='n').grid(row=next_row+1)
    Xvar = tk.IntVar()
    for i,x in enumerate(h):
        x_row=i+2
        tk.Radiobutton(variables_frame,text=x,variable=Xvar, value=i).grid(row=next_row+x_row)
    #Yvar
    tk.Label(variables_frame,text="Select Y-axis variable to plot",anchor='n').grid(row=next_row+1,column=1)
    Yvar = tk.IntVar()
    for j,x in enumerate(h):
        y_row=j+2
        tk.Radiobutton(variables_frame,text=x,variable=Yvar, value=j).grid(row=next_row+y_row,column=1)


mypath = os.getcwd()
# -------------------------------------------------------
# ------- Reading and spliting the main log file --------
# -------------------------------------------------------
flag_loglammps=0
# --- file selection ---
try :
    in_file = sys.argv[1]
except:
    print("Usage : plot_general.py file_name")
    quit()

if in_file == "log.lammps":
    flag_loglammps=1
    data = open("log.lammps",'r').readlines()  # select log.lammps as main log file if it exists
      
    # --- spliting into "out.log.*" files ---
    j = 0
    flag = 0
    flag_out = 0
    step = ""
    for lines in data:
        lines = lines.strip()
        if lines.startswith("run:"):
            step = "."+lines.strip().strip("run: ")
        elif lines.startswith("Memory usage per processor"):
          if len(str(j)) == 1:
              J = "0"+str(j)
          else:
              J = str(j)
          out = open("out.log.%s%s-plt-tmp"%(J,step),'w')
          flag_out = 1
          flag = 1
        elif flag and lines.startswith("Loop") or lines.startswith("kill"):
            flag = 0
            j+=1
            out.close()
            step = ""
        elif flag:
            out.write(lines+'\n')
    if flag_out:
        out.close()
    outlogfiles = [ f for f in os.listdir(mypath) if os.path.isfile(os.path.join(mypath,f)) and f.startswith("out.log") and f.endswith("plt-tmp")] # list of "out.log.*" files
else:
    outlogfiles = [in_file]
    pass
  




###################
### MAIN script ###
###################

root=tk.Tk()
# Positioning frames
button_frame = tk.Frame(root,heigh=5)
button_frame.pack(side='right')

source_frame = tk.Frame(root)
source_frame.pack()

labels_frame = tk.Frame(root)
labels_frame.pack(side="bottom")

variables_frame = tk.Frame(root)
variables_frame.pack(side="bottom")

# Filling frames
tk.Label(source_frame,text="Select data source",anchor='n').grid(row=0)
Fvar = tk.IntVar()
for i,x in enumerate(outlogfiles):
    next_row=i+1
    tk.Radiobutton(source_frame,text=x,variable=Fvar, value=i, command = process_file).grid(row=next_row)
    if not flag_loglammps:
        skipline = tk.Entry(source_frame,width=5,justify="center")
        skipline.grid(row=next_row,column=1)
        skipline.insert(0,0)
    else:
        skipline=0
next_row+=1  
ttk.Separator(source_frame,orient='horizontal').grid(row=next_row,columnspan=2,sticky="ew") 

# Insert X,Y label text field
ttk.Separator(labels_frame,orient='horizontal').grid(columnspan=2,sticky="ew") 
xlabel = tk.Entry(labels_frame,width=50,justify="center")
xlabel.grid(row=next_row+1,column=0,columnspan=2)
xlabel.insert(0,"X-axis label")
ylabel = tk.Entry(labels_frame,width=50,justify="center")
ylabel.grid(row=next_row+2,column=0,columnspan=2)
ylabel.insert(0,"Y-axis label")

close = tk.Button(button_frame, text = "Close",width=10, command = close_windows)
close.grid(row=1)
plot = tk.Button(button_frame, text = "Plot",width=10, command = plot_button)
plot.grid(row=0)

# show GUI
root.mainloop()


# Cleaning tmp files
if flag_loglammps:
    for f in outlogfiles:
        os.remove(f)



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

import numpy as np

###

from _matscipy import count_islands, count_segments, distance_map

###

# Stencils for determining nearest-neighbor relationships on a square grid
nn_stencil = [(1,0),(0,1),(-1,0),(0,-1)]
nnn_stencil = [(1,1),(1,0),(1,-1),(0,-1),(-1,-1),(-1,0),(-1,1),(0,1)]

###

def outer_perimeter(c, stencil=nn_stencil):
    """
    Return a map where the outer perimeter is marked with the patch number.

    Parameters
    ----------
    c : array_like
        2D map of islands.

    Returns
    -------
    c : array
        2D map of island perimeters.
    """
    
    c_nearby = np.zeros_like(c, dtype=bool)
    for dx, dy in stencil:
        tmp = c.copy()
        if dx != 0:
            tmp = np.roll(tmp, dx, 0)
        if dy != 0:
            tmp = np.roll(tmp, dy, 1)
        c_nearby = np.logical_or(c_nearby, tmp)
    return np.logical_and(np.logical_not(c), c_nearby)


def inner_perimeter(patch_ids, stencil=nn_stencil):
    """
    Return a map where the inner perimeter is marked with the patch number

    Parameters
    ----------
    c : array_like
        2D map of islands.

    Returns
    -------
    c : array
        2D map of island perimeters.
    """
    
    c = outer_perimeter(patch_ids == 0, stencil)
    return c*patch_ids


def island_areas(island_ids):
    """
    Return a list containing island areas for each island number

    Parameters
    ----------
    c : array_like
        2D map of island ids.
  
    Returns
    -------
    areas : array
        Array containing the area for each island id.
    """
    
    return np.bincount(island_ids.reshape((-1,)))[1:]



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

import itertools

import numpy as np

from ase.data import atomic_numbers

import _matscipy
from _matscipy import first_neighbours

###

def mic(dr, cell, pbc=None):
    """
    Apply minimum image convention to an array of distance vectors.

    Parameters
    ----------
    dr : array_like
        Array of distance vectors.
    cell : array_like
        Simulation cell.
    pbc : array_like, optional
        Periodic boundary conditions in x-, y- and z-direction. Default is to
        assume periodic boundaries in all directions.

    Returns
    -------
    dr : array
        Array of distance vectors, wrapped according to the minimum image
        convention.
    """
    # Check where distance larger than 1/2 cell. Particles have crossed
    # periodic boundaries then and need to be unwrapped.
    rec = np.linalg.inv(cell)
    if pbc is not None:
        rec *= np.array(pbc, dtype=int).reshape(3,1)
    dri = np.round(np.dot(dr, rec))

    # Unwrap
    return dr - np.dot(dri, cell)


def neighbour_list(quantities, a, cutoff):
    """
    Compute a neighbor list for an atomic configuration. Atoms outside periodic
    boundaries are mapped into the box. Atoms outside nonperiodic boundaries
    are included in the neighbor list but complexity of neighbor list search
    for those can become n^2.

    The neighbor list is sorted by first atom index 'i', but not by second 
    atom index 'j'.

    Parameters
    ----------
    quantities : str
        Quantities to compute by the neighbor list algorithm. Each character
        in this string defines a quantity. They are returned in a tuple of
        the same order. Possible quantities are
            'i' : first atom index
            'j' : second atom index
            'd' : absolute distance
            'D' : distance vector
            'S' : shift vector (number of cell boundaries crossed by the bond
                  between atom i and j). With the shift vector S, the
                  distances D between atoms can be computed from:
                  D = a.positions[j]-a.positions[i]+S.dot(a.cell)
    a : ase.Atoms
        Atomic configuration.
    cutoff : float or dict
        Cutoff for neighbor search. It can be
            - A single float: This is a global cutoff for all elements.
            - A dictionary: This specifies cutoff values for element
              pairs. Specification accepts element numbers of symbols.
              Example: {(1, 6): 1.1, (1, 1): 1.0, ('C', 'C'): 1.85}
            - A list/array with a per atom value: This specifies the radius of
              an atomic sphere for each atoms. If spheres overlap, atoms are
              within each others neighborhood.

    Returns
    -------
    i, j, ... : array
        Tuple with arrays for each quantity specified above. Indices in `i`
        are returned in ascending order 0..len(a), but the order of (i,j)
        pairs is not guaranteed.

    Examples
    --------
    Examples assume Atoms object *a* and numpy imported as *np*.
    1. Coordination counting:
        i = neighbor_list('i', a, 1.85)
        coord = np.bincount(i)

    2. Coordination counting with different cutoffs for each pair of species
        i = neighbor_list('i', a,
                           {('H', 'H'): 1.1, ('C', 'H'): 1.3, ('C', 'C'): 1.85})
        coord = np.bincount(i)

    3. Pair distribution function:
        d = neighbor_list('d', a, 10.00)
        h, bin_edges = np.histogram(d, bins=100)
        pdf = h/(4*np.pi/3*(bin_edges[1:]**3 - bin_edges[:-1]**3)) * a.get_volume()/len(a)

    4. Pair potential:
        i, j, d, D = neighbor_list('ijdD', a, 5.0)
        energy = (-C/d**6).sum()
        pair_forces = (6*C/d**5  * (D/d).T).T
        forces_x = np.bincount(j, weights=pair_forces[:, 0], minlength=len(a)) - \
                   np.bincount(i, weights=pair_forces[:, 0], minlength=len(a))
        forces_y = np.bincount(j, weights=pair_forces[:, 1], minlength=len(a)) - \
                   np.bincount(i, weights=pair_forces[:, 1], minlength=len(a))
        forces_z = np.bincount(j, weights=pair_forces[:, 2], minlength=len(a)) - \
                   np.bincount(i, weights=pair_forces[:, 2], minlength=len(a))

    5. Dynamical matrix for a pair potential stored in a block sparse format:
        from scipy.sparse import bsr_matrix
        i, j, dr, abs_dr = neighbor_list('ijDd', atoms)
        energy = (dr.T / abs_dr).T
        dynmat = -(dde * (energy.reshape(-1, 3, 1) * energy.reshape(-1, 1, 3)).T).T \
                 -(de / abs_dr * (np.eye(3, dtype=energy.dtype) - \
                   (energy.reshape(-1, 3, 1) * energy.reshape(-1, 1, 3))).T).T
        dynmat_bsr = bsr_matrix((dynmat, j, first_i), shape=(3*len(a), 3*len(a)))

        dynmat_diag = np.empty((len(a), 3, 3))
        for x in range(3):
            for y in range(3):
                dynmat_diag[:, x, y] = -np.bincount(i, weights=dynmat[:, x, y])

        dynmat_bsr += bsr_matrix((dynmat_diag, np.arange(len(a)),
                                  np.arange(len(a) + 1)),
                                 shape=(3 * len(a), 3 * len(a)))


  i_n, j_n, dr_nc, abs_dr_n = neighbour_list('ijDd', atoms, dict)
e_nc = (dr_nc.T/abs_dr_n).T
    D_ncc = -(dde_n * (e_nc.reshape(-1,3,1) * e_nc.reshape(-1,1,3)).T).T
    D_ncc += -(de_n/abs_dr_n * (np.eye(3, dtype=e_nc.dtype) - (e_nc.reshape(-1,3,1) * e_nc.reshape(-1,1,3))).T).T

    D = bsr_matrix((D_ncc, j_n, first_i), shape=(3*nat,3*nat))

    Ddiag_icc = np.empty((nat,3,3))
    for x in range(3):
        for y in range(3):
            Ddiag_icc[:,x,y] = -np.bincount(i_n, weights = D_ncc[:,x,y])

    D += bsr_matrix((Ddiag_icc,np.arange(nat),np.arange(nat+1)), shape=(3*nat,3*nat))

    return D
    """

    if isinstance(cutoff, dict):
        maxel = np.max(a.numbers)
        _cutoff = np.zeros([maxel+1, maxel+1], dtype=float)
        for (el1, el2), c in cutoff.items():
            try:
                el1 = atomic_numbers[el1]
            except:
                pass
            try:
                el2 = atomic_numbers[el2]
            except:
                pass
            if el1 < maxel+1 and el2 < maxel+1:
                _cutoff[el1, el2] = c
                _cutoff[el2, el1] = c
    else:
        _cutoff = cutoff

    return _matscipy.neighbour_list(quantities, a.cell,
                                    np.linalg.inv(a.cell.T), a.pbc,
                                    a.positions, _cutoff,
                                    a.numbers.astype(np.int32))


import os
import glob
import sys
import re

from numpy.distutils.core import setup, Extension

from numpy.distutils.command.build_ext import build_ext
from numpy.distutils import log
from distutils.dep_util import newer_group
from numpy.distutils.misc_util import filter_sources, has_f_sources, \
     has_cxx_sources, get_ext_source_files, \
     get_numpy_include_dirs, is_sequence, get_build_architecture, \
     msvc_version

import versioneer

# custom compilation options for various compilers
# entry with key None gives default vaalues always used
copt =  {
          None: []
        }

cxxopt = {
          None: ['-std=c++0x']
         }

lopt =  {
        }

version = versioneer.get_version()
_version_short = re.findall('\d+\.\d+\.\d+', version)
if len(_version_short) > 0:
    version_short = _version_short[0]
else:
    version_short = 'master'
download_url = "https://github.com/libAtoms/matscipy/archive/%s.tar.gz" % version_short

scripts = []

# subclass build_ext so that we can override build_extension()
# to supply different flags to the C and C++ compiler
class build_ext_subclass(build_ext):
    def build_extension(self, ext):
        sources = ext.sources
        if sources is None or not is_sequence(sources):
            raise DistutilsSetupError(
                ("in 'ext_modules' option (extension '%s'), " +
                 "'sources' must be present and must be " +
                 "a list of source filenames") % ext.name)
        sources = list(sources)

        if not sources:
            return

        fullname = self.get_ext_fullname(ext.name)
        if self.inplace:
            modpath = fullname.split('.')
            package = '.'.join(modpath[0:-1])
            base = modpath[-1]
            build_py = self.get_finalized_command('build_py')
            package_dir = build_py.get_package_dir(package)
            ext_filename = os.path.join(package_dir,
                                        self.get_ext_filename(base))
        else:
            ext_filename = os.path.join(self.build_lib,
                                        self.get_ext_filename(fullname))
        depends = sources + ext.depends

        if not (self.force or newer_group(depends, ext_filename, 'newer')):
            log.debug("skipping '%s' extension (up-to-date)", ext.name)
            return
        else:
            log.info("building '%s' extension", ext.name)

        extra_args = ext.extra_compile_args or []
        cxx_extra_args = ext.extra_compile_args or []
        extra_link_args = ext.extra_link_args or []

        c = os.path.basename(self.compiler.compiler[0])
        cxx = os.path.basename(self.compiler.compiler_cxx[0])
        if None in copt:
            extra_args += copt[None]
        if c in copt:
            extra_args += copt[c]
        if None in cxxopt:
            cxx_extra_args += cxxopt[None]
        if cxx in cxxopt:
            cxx_extra_args += cxxopt[cxx]
        if None in lopt:
            extra_link_args += lopt[None]
        if c in lopt:
            extra_link_args += lopt[c]
        if cxx in lopt:
            extra_link_args += lopt[cxx]

        macros = ext.define_macros[:]
        for undef in ext.undef_macros:
            macros.append((undef,))

        c_sources, cxx_sources, f_sources, fmodule_sources = \
                   filter_sources(ext.sources)

        if self.compiler.compiler_type=='msvc':
            if cxx_sources:
                # Needed to compile kiva.agg._agg extension.
                cxx_extra_args.append('/Zm1000')
            # this hack works around the msvc compiler attributes
            # problem, msvc uses its own convention :(
            c_sources += cxx_sources
            cxx_sources = []

        # Set Fortran/C++ compilers for compilation and linking.
        if ext.language=='f90':
            fcompiler = self._f90_compiler
        elif ext.language=='f77':
            fcompiler = self._f77_compiler
        else: # in case ext.language is c++, for instance
            fcompiler = self._f90_compiler or self._f77_compiler
        if fcompiler is not None:
            fcompiler.extra_f77_compile_args = (ext.extra_f77_compile_args or []) if hasattr(ext, 'extra_f77_compile_args') else []
            fcompiler.extra_f90_compile_args = (ext.extra_f90_compile_args or []) if hasattr(ext, 'extra_f90_compile_args') else []
        cxx_compiler = self._cxx_compiler

        # check for the availability of required compilers
        if cxx_sources and cxx_compiler is None:
            raise DistutilsError("extension %r has C++ sources" \
                  "but no C++ compiler found" % (ext.name))
        if (f_sources or fmodule_sources) and fcompiler is None:
            raise DistutilsError("extension %r has Fortran sources " \
                  "but no Fortran compiler found" % (ext.name))
        if ext.language in ['f77', 'f90'] and fcompiler is None:
            self.warn("extension %r has Fortran libraries " \
                  "but no Fortran linker found, using default linker" % (ext.name))
        if ext.language=='c++' and cxx_compiler is None:
            self.warn("extension %r has C++ libraries " \
                  "but no C++ linker found, using default linker" % (ext.name))

        kws = {'depends':ext.depends}
        output_dir = self.build_temp

        include_dirs = ext.include_dirs + get_numpy_include_dirs()

        c_objects = []
        if c_sources:
            log.info("compiling C sources with arguments %r", extra_args)
            c_objects = self.compiler.compile(c_sources,
                                              output_dir=output_dir,
                                              macros=macros,
                                              include_dirs=include_dirs,
                                              debug=self.debug,
                                              extra_postargs=extra_args,
                                              **kws)

        if cxx_sources:
            log.info("compiling C++ sources with arguments %r", cxx_extra_args)
            c_objects += cxx_compiler.compile(cxx_sources,
                                              output_dir=output_dir,
                                              macros=macros,
                                              include_dirs=include_dirs,
                                              debug=self.debug,
                                              extra_postargs=cxx_extra_args,
                                              **kws)

        extra_postargs = []
        f_objects = []
        if fmodule_sources:
            log.info("compiling Fortran 90 module sources")
            module_dirs = ext.module_dirs[:]
            module_build_dir = os.path.join(
                self.build_temp, os.path.dirname(
                    self.get_ext_filename(fullname)))

            self.mkpath(module_build_dir)
            if fcompiler.module_dir_switch is None:
                existing_modules = glob('*.mod')
            extra_postargs += fcompiler.module_options(
                module_dirs, module_build_dir)
            f_objects += fcompiler.compile(fmodule_sources,
                                           output_dir=self.build_temp,
                                           macros=macros,
                                           include_dirs=include_dirs,
                                           debug=self.debug,
                                           extra_postargs=extra_postargs,
                                           depends=ext.depends)

            if fcompiler.module_dir_switch is None:
                for f in glob('*.mod'):
                    if f in existing_modules:
                        continue
                    t = os.path.join(module_build_dir, f)
                    if os.path.abspath(f)==os.path.abspath(t):
                        continue
                    if os.path.isfile(t):
                        os.remove(t)
                    try:
                        self.move_file(f, module_build_dir)
                    except DistutilsFileError:
                        log.warn('failed to move %r to %r' %
                                 (f, module_build_dir))
        if f_sources:
            log.info("compiling Fortran sources")
            f_objects += fcompiler.compile(f_sources,
                                           output_dir=self.build_temp,
                                           macros=macros,
                                           include_dirs=include_dirs,
                                           debug=self.debug,
                                           extra_postargs=extra_postargs,
                                           depends=ext.depends)

        objects = c_objects + f_objects

        if ext.extra_objects:
            objects.extend(ext.extra_objects)
        libraries = self.get_libraries(ext)[:]
        library_dirs = ext.library_dirs[:]

        linker = self.compiler.link_shared_object
        # Always use system linker when using MSVC compiler.
        if self.compiler.compiler_type=='msvc':
            # expand libraries with fcompiler libraries as we are
            # not using fcompiler linker
            self._libs_with_msvc_and_fortran(fcompiler, libraries, library_dirs)

        elif ext.language in ['f77', 'f90'] and fcompiler is not None:
            linker = fcompiler.link_shared_object
        if ext.language=='c++' and cxx_compiler is not None:
            linker = cxx_compiler.link_shared_object

        if sys.version[:3]>='2.3':
            kws = {'target_lang':ext.language}
        else:
            kws = {}

        linker(objects, ext_filename,
               libraries=libraries,
               library_dirs=library_dirs,
               runtime_library_dirs=ext.runtime_library_dirs,
               extra_postargs=extra_link_args,
               export_symbols=self.get_export_symbols(ext),
               debug=self.debug,
               build_temp=self.build_temp,**kws)


cmdclass = versioneer.get_cmdclass()
cmdclass.update({'build_ext': build_ext_subclass})
setup(name='matscipy',
      version=version,
      cmdclass=cmdclass,
      description='Generic Python Materials Science tools',
      maintainer='James Kermode & Lars Pastewka',
      maintainer_email='james.kermode@gmail.com',
      license='LGPLv2.1+',
      package_dir={'matscipy': 'matscipy'},
      packages=['matscipy', 'matscipy.fracture_mechanics',
                'matscipy.contact_mechanics', 'matscipy.calculators',
                'matscipy.calculators.eam','matscipy.io','matscipy.tool',
                'matscipy.calculators.pair_potential',
                'matscipy.calculators.mcfm',
                'matscipy.calculators.mcfm.mcfm_parallel',
                'matscipy.calculators.mcfm.neighbour_list_mcfm',
                'matscipy.calculators.mcfm.qm_cluster_tools'],
      scripts=scripts,
      ext_modules=[
        Extension(
            '_matscipy',
            ['c/tools.c',
             'c/angle_distribution.c',
             'c/neighbours.c',
             'c/islands.cpp',
             'c/ring_statistics.cpp',
             'c/matscipymodule.c'],
            )
        ],
      download_url=download_url,
      url="https://github.com/libAtoms/matscipy"
      )



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014-2017) James Kermode, Warwick University
#                       Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

import numpy as np
from matscipy.neighbours import mic

###

class RemoveDrift:
    """
    Remove drift of the center of mass motion of an atomic system that may
    occur in periodic simulations.
    """

    def __init__(self, traj):
        self.traj = traj

        # Stores the global displacement vectors (shift vectors) required to
        # move all positions back to original.
        self.shifts = [np.zeros(3)]

    def _fill_shifts_upto(self, i):
        # Iterate up to frame i the full trajectory first and generate a list
        # of displacement vectors.
        while len(self.shifts) <= i:
            j = len(self.shifts)
            a0 = self.traj[j-1]
            a1 = self.traj[j]
            s0 = a0.get_scaled_positions()%1.0
            s1 = a1.get_scaled_positions()%1.0
            sdisps = mic(s1-s0, np.eye(3), pbc=a0.pbc)
            self.shifts += [self.shifts[-1]+sdisps.mean(axis=0)]

    def __getitem__(self, i=-1):
        if i < 0:
            i = len(self) + i
            if i < 0 or i >= len(self):
                raise IndexError('Trajectory index out of range.')

        self._fill_shifts_upto(i)

        a = self.traj[i]
        a.set_scaled_positions(a.get_scaled_positions()%1.0-self.shifts[i])

        return a


    def __len__(self):
        return len(self.traj)




from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np

from math import sqrt
from .neighbour_list_base import NeighbourListBase
from ....neighbours import neighbour_list as mspy_nl


class NeighbourListMCFM(NeighbourListBase):
    """Neighbor list object. Wrapper aroud matscipy.neighbour_list

    atoms: ase.Atoms
        Atomic configuration.
    cutoffs: float or dict
        Cutoff for neighbour search. If single float is given, a global cutoff
        is used for all elements. A dictionary specifies cutoff for element
        pairs. Specification accepts element numbers of symbols.
        Example: {(1, 6): 1.1, (1, 1): 1.0, ('C', 'C'): 1.85}
    skin: float
        If no atom has moved more than the skin-distance since the
        last call to the ``update()`` method, then the neighbor list
        can be reused.  This will save some expensive rebuilds of
        the list, but extra neighbors outside the cutoff will be
        returned.
    hysteretic_break_factor: float
        If atoms are connected, the link will break only of they move apart
        further than cutoff * hysteretic_break_factor

    """

    def __init__(self, atoms, cutoffs, skin=0.3, hysteretic_break_factor=1):

        self.cutoffs = cutoffs.copy()
        self.cutoffs_hysteretic = cutoffs.copy()
        if hysteretic_break_factor > 1:
            self.do_hysteretic = True
            for key in self.cutoffs_hysteretic:
                self.cutoffs_hysteretic[key] *= hysteretic_break_factor
        else:
            self.do_hysteretic = False

        self.skin = skin
        self.nupdates = 0

        # Additional data
        self.neighbours = [np.zeros(0) for idx in range(len(atoms))]
        self.old_neighbours = [[] for idx in range(len(atoms))]

    def update(self, atoms):
        """Make sure the list is up to date. If clled for the first
        time, build the list

        Parameters
        ----------
        atoms : ase.Atoms
            atoms to initialize the list from

        Returns
        -------
        bool
            True of the update was sucesfull
        """
        if self.nupdates == 0:
            self.do_update(atoms)
            return True
        elif ((self.pbc != atoms.get_pbc()).any() or
              (self.cell != atoms.get_cell()).any() or
              ((self.positions - atoms.get_positions())**2).sum(1).max() >
                self.skin**2):
            self.do_update(atoms)
            return True

        return False

    def do_update(self, atoms):
        """Build the neighbour list based on pairwise distances.

        Parameters
        ----------
        atoms : ase.Atoms
            atoms to initialize the list from

        Raises
        ------
        ValueError
            Must specify cutoff radii for all atoms
        """
        self.positions = atoms.get_positions()
        self.pbc = atoms.get_pbc()
        self.cell = atoms.get_cell()

        shorti, shortj = mspy_nl(str("ij"), atoms, self.cutoffs)

        new_neighbours = [[] for idx in range(len(atoms))]
        for idx in range(len(shorti)):
            new_neighbours[shorti[idx]].append(shortj[idx])

        if self.do_hysteretic:
            longi, longj = mspy_nl(str("ij"), atoms, self.cutoffs_hysteretic)

            for idx in range(len(longi)):
                # Split for profiling
                previously_connected = longj[idx] in self.old_neighbours[longi[idx]]
                not_added = longj[idx] not in new_neighbours[longi[idx]]
                if previously_connected and not_added:
                    new_neighbours[longi[idx]].append(longj[idx])

            self.old_neighbours = new_neighbours

        for idx in range(len(new_neighbours)):
            self.neighbours[idx] = np.asarray(list(new_neighbours[idx]))

        self.nupdates += 1

    def get_neighbours(self, a):
        """Return neighbors of atom number a.

        A list of indices to neighboring atoms is
        returned.

        Parameters
        ----------
        a : int
            atomic index


        Returns
        -------
        np.array
            array of neighbouring indices

        Raises
        ------
        RuntimeError
            Must update the list at least once!

        """

        if self.nupdates == 0:
            raise RuntimeError("Must update the list at least once!")

        return self.neighbours[a]



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

import itertools

import numpy as np

from matscipy.neighbours import neighbour_list
import _matscipy 

###

def ring_statistics(a, cutoff, maxlength=-1):
    """
    Compute number of shortest path rings in sample.
    See: D.S. Franzblau, Phys. Rev. B 44, 4925 (1991)

    Parameters
    ----------
    a : ase.Atoms
        Atomic configuration.
    cutoff : float
        Cutoff for neighbor counting.
    maxlength : float, optional
        Maximum ring length. Search for rings will stop at this length. This
        is useful to speed up calculations for large systems.

    Returns
    -------
    ringstat : array
        Array with number of shortest path rings.
    """
    i, j, r = neighbour_list('ijD', a, cutoff)
    d = _matscipy.distances_on_graph(i, j)
    return _matscipy.find_sp_rings(i, j, r, d, maxlength)



from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np


class NeighbourListBase(object):
    """Interface for the neighbour list.
    mcfm module can use any neighbour list object as long
    as it provides the implementation of the two routines below.
    """

    def update(self, atoms):
        """Make sure the list is up to date. If clled for the first
        time, build the list

        Parameters
        ----------
        atoms : ase.Atoms
            atoms to initialize the list from

        Returns
        -------
        bool
            True of the update was sucesfull
        """
        raise NotImplementedError("Must implement this function!")

    def get_neighbours(self, a):
        """Return neighbors of atom number a.

        A list of indices to neighboring atoms is
        returned.

        Parameters
        ----------
        a : int
            atomic index


        Returns
        -------
        np.array
            array of neighbouring indices

        """
        raise NotImplementedError("Must implement this function!")
